#ifndef SENSOR_GPS_H_
#define SENSOR_GPS_H_

//wolf
#include "core/sensor/sensor_base.h"

// std

namespace wolf {

WOLF_STRUCT_PTR_TYPEDEFS(IntrinsicsGnss);
WOLF_PTR_TYPEDEFS(SensorGnss);

struct IntrinsicsGnss : public IntrinsicsBase
{
        // add GNSS parameters here
        bool extrinsics_fixed_ = true;
        bool roll_fixed_ =true;
        bool pitch_fixed_=true;
        bool yaw_fixed_  =false;
        bool translation_fixed_ = false;
        virtual ~IntrinsicsGnss() = default;
        IntrinsicsGnss() = default;
        IntrinsicsGnss(std::string _unique_name, const paramsServer& _server):
            IntrinsicsBase(_unique_name, _server)
        {
            extrinsics_fixed_ = _server.getParam<bool>(_unique_name + "/extrinsics_fixed");
            roll_fixed_ = _server.getParam<bool>(_unique_name + "/roll_fixed");
            pitch_fixed_ = _server.getParam<bool>(_unique_name + "/pitch_fixed");
            yaw_fixed_ = _server.getParam<bool>(_unique_name + "/yaw_fixed");
            translation_fixed_ = _server.getParam<bool>(_unique_name + "/translation_fixed");
        }
  std::string print()
  {
      return "\n" + IntrinsicsBase::print()
          + "extrinsics_fixed: " + std::to_string(extrinsics_fixed_) + "\n"
          + "roll_fixed: " + std::to_string(roll_fixed_) + "\n"
          + "pitch_fixed: " + std::to_string(pitch_fixed_) + "\n"
          + "yaw_fixed: " + std::to_string(yaw_fixed_) + "\n"
          + "translation_fixed: " + std::to_string(translation_fixed_) + "\n";
  }
};

class SensorGnss : public SensorBase
{
    protected:
        IntrinsicsGnssPtr params_;
        bool ENU_defined_, ENU_MAP_initialized_;
        Eigen::Matrix3s R_ENU_ECEF_;
        Eigen::Vector3s t_ENU_ECEF_;

    public:
        SensorGnss(StateBlockPtr _p_ptr, StateBlockPtr _bias_ptr, IntrinsicsGnssPtr params);
        SensorGnss(StateBlockPtr _p_ptr, StateBlockPtr _bias_ptr, IntrinsicsGnssPtr params, StateBlockPtr _t_ENU_MAP_ptr, StateBlockPtr _roll_ENU_MAP_ptr, StateBlockPtr _pitch_ENU_MAP_ptr, StateBlockPtr _yaw_ENU_MAP_ptr);

        virtual ~SensorGnss();

        // Gets
        StateBlockPtr getEnuMapTranslation() const;
        StateBlockPtr getEnuMapRoll() const;
        StateBlockPtr getEnuMapPitch() const;
        StateBlockPtr getEnuMapYaw() const;
        const Eigen::Matrix3s& getREnuEcef() const;
        const Eigen::Vector3s& gettEnuEcef() const;
        Eigen::Matrix3s getREnuMap() const;
        Eigen::Vector3s gettEnuMap() const;
        bool isEnuDefined() const;
        bool isEnuMapInitialized() const;

        // Sets
        void setEnuMapTranslationState(const Eigen::Vector3s& t_ENU_MAP);
        void setEnuMapRollState(const Scalar& roll_ENU_MAP);
        void setEnuMapPitchState(const Scalar& pitch_ENU_MAP);
        void setEnuMapYawState(const Scalar& yaw_ENU_MAP);
        void setEcefEnu(const Eigen::Vector3s& _t_ECEF_ENU);
        void setEnuEcef(const Eigen::Matrix3s& _R_ENU_ECEF, const Eigen::Vector3s& _t_ENU_ECEF);

        // compute
        template<typename T>
        Eigen::Matrix<T,3,3> computeREnuMap(const T& _r,const T& _p,const T& _y) const;
        void computeEnuEcef(const Eigen::Vector3s& _t_ECEF_ENU, Eigen::Matrix3s& R_ENU_ECEF, Eigen::Vector3s& t_ENU_ECEF) const;
        void initializeEnuMap(const Eigen::VectorXs& _pose_MAP_frameENU, const Eigen::Vector3s& _t_ECEF_antenaENU,
                              const Eigen::VectorXs& _pose_MAP_frame2, const Eigen::Vector3s& _t_ECEF_antena2);
        void initializeEnuMapYaw(const Eigen::VectorXs& _pose_MAP_frame1, const Eigen::VectorXs& _pose_MAP_frame2,
                                 const Eigen::Vector3s& _v_ECEF_antena1_antena2);

    public:
        static SensorBasePtr create(const std::string& _unique_name, const Eigen::VectorXs& _extrinsics_p, const IntrinsicsBasePtr _intrinsics);


};

inline bool SensorGnss::isEnuDefined() const
{
    return ENU_defined_;
}

inline bool SensorGnss::isEnuMapInitialized() const
{
    return ENU_MAP_initialized_;
}

inline StateBlockPtr SensorGnss::getEnuMapTranslation() const
{
    return getStateBlockPtrStatic(3);
}

inline StateBlockPtr SensorGnss::getEnuMapRoll() const
{
    return getStateBlockPtrStatic(4);
}

inline StateBlockPtr SensorGnss::getEnuMapPitch() const
{
    return getStateBlockPtrStatic(5);
}

inline StateBlockPtr SensorGnss::getEnuMapYaw() const
{
    return getStateBlockPtrStatic(6);
}

inline const Eigen::Matrix3s& SensorGnss::getREnuEcef() const
{
    return R_ENU_ECEF_;
}

inline const Eigen::Vector3s& SensorGnss::gettEnuEcef() const
{
    return t_ENU_ECEF_;
}

template<typename T>
inline Eigen::Matrix<T,3,3> SensorGnss::computeREnuMap(const T& _r,const T& _p,const T& _y) const
{
    Eigen::Matrix<T,3,3> R_ENU_MAP = (Eigen::AngleAxis<T>(_r, Eigen::Matrix<T,3,1>::UnitX())
                                    * Eigen::AngleAxis<T>(_p, Eigen::Matrix<T,3,1>::UnitY())
                                    * Eigen::AngleAxis<T>(_y, Eigen::Matrix<T,3,1>::UnitZ())).toRotationMatrix();

    return R_ENU_MAP;

    // TODO: store R's when T=Scalar
    //Eigen::Matrix<T,3,3> R_roll_T, R_pitch_T, R_yaw_T;
    //if (!getRollState()->isFixed())
    //    *R_roll_ptr_ = Eigen::AngleAxis<Scalar>(0.25*M_PI, Eigen::Vector3s::UnitX());
}

} // namespace wolf

#endif //SENSOR_GPS_H_
