#include "gnss/factor/factor_gnss_fix_2D.h"
#include "gnss/factor/factor_gnss_fix_3D.h"
#include "gnss/feature/feature_gnss_fix.h"
#include "gnss/processor/processor_gnss_fix.h"
#include "gnss/capture/capture_gnss_fix.h"

namespace wolf
{

ProcessorGnssFix::ProcessorGnssFix(ProcessorParamsGnssFixPtr _params_gnss, SensorGnssPtr _sensor_gnss) :
        ProcessorBase("GNSS FIX", _params_gnss),
        sensor_gnss_(_sensor_gnss),
        params_gnss_(_params_gnss),
        first_capture_(nullptr)
{
    //
}

ProcessorGnssFix::~ProcessorGnssFix()
{
    //
}

void ProcessorGnssFix::processCapture(CaptureBasePtr _capture)
{
    // TODO: keep captures in a buffer and deal with KFpacks
    //WOLF_DEBUG("ProcessorGnssFix::process()");
    incoming_capture_ = std::static_pointer_cast<CaptureGnssFix>(_capture);

    FrameBasePtr new_frame = nullptr;
    FactorBasePtr new_fac = nullptr;

    // ALREADY CREATED KF
    PackKeyFramePtr KF_pack = buffer_pack_kf_.selectPack( incoming_capture_, params_gnss_->time_tolerance);
    if (KF_pack && KF_pack->key_frame != last_KF_)
    {
        WOLF_DEBUG( "PR ", getName()," - capture ", incoming_capture_->id(), " appended to existing KF " , KF_pack->key_frame->id() , " TS: ", KF_pack->key_frame->getTimeStamp());
        new_frame = KF_pack->key_frame;
    }
    // MAKE KF
    else if (voteForKeyFrame() && permittedKeyFrame())
    {
        new_frame = getProblem()->emplaceFrame(KEY, incoming_capture_->getTimeStamp());
        getProblem()->keyFrameCallback(new_frame, shared_from_this(), params_gnss_->time_tolerance);
        WOLF_DEBUG( "PR ", getName()," - capture ", incoming_capture_->id(), " appended to new KF " , new_frame->id() , " TS: ", new_frame->getTimeStamp());
    }

    // ESTABLISH FACTOR
    if (new_frame)
    {
        // LINK CAPTURE
        incoming_capture_->link(new_frame); // Add incoming Capture to the new Frame

        // EMPLACE FEATURES
        WOLF_DEBUG( "PR ", getName()," - emplacing the feature...");
        auto ftr = FeatureBase::emplace<FeatureGnssFix>(incoming_capture_, incoming_capture_->getData(),incoming_capture_->getDataCovariance());

        // EMPLACE FACTOR
        new_fac = emplaceFactor(ftr);

        // outlier rejection
        if (sensor_gnss_->isEnuDefined() && sensor_gnss_->isEnuMapInitialized())
            if (rejectOutlier(new_fac))
                new_fac = nullptr;

        // store last KF
        if (new_fac)
            last_KF_= new_frame;
    }

    // SET ECEF_ENU IF:
    //      1 - not initialized
    //      2 - factor established
    if (!sensor_gnss_->isEnuDefined() && new_fac != nullptr)
    {
        WOLF_DEBUG("setting ecef enu");
        sensor_gnss_->setEcefEnu(incoming_capture_->getData());

        // Store the first capture that established a factor
        first_capture_ = incoming_capture_;
    }

    // INITIALIZE ENU_MAP IF 4 NECESSARY CONDITIONS:
    //      1 - ENU-ECEF defined
    //      2 - not initialized
    //      3 - current capture in key-frame with factor
    //      4 - two factors established (first and current) in frames separated enough ( > enu_map_init_dist_min)
    if ( sensor_gnss_->isEnuDefined() &&
        !sensor_gnss_->isEnuMapInitialized() &&
         new_fac != nullptr &&
         first_capture_ != nullptr && first_capture_->getFrame() != nullptr && incoming_capture_->getFrame() != nullptr &&
         (first_capture_->getFrame()->getState()-incoming_capture_->getFrame()->getState()).norm() > params_gnss_->enu_map_init_dist_min)
    {
        WOLF_DEBUG("initializing enu map");
        sensor_gnss_->initializeEnuMap(first_capture_->getFrame()->getState(),      first_capture_->getData(),
                                       incoming_capture_->getFrame()->getState(),   incoming_capture_->getData());
    }
}

FactorBasePtr ProcessorGnssFix::emplaceFactor(FeatureBasePtr& ftr)
{
    // CREATE CONSTRAINT --------------------
    //WOLF_DEBUG("creating the factor...");
    // 2D
    if (getProblem()->getDim() == 2)
        return FactorBase::emplace<FactorGnssFix2D>(ftr, ftr, sensor_gnss_, shared_from_this(), false, FAC_ACTIVE);
    // 3D
    else
        return FactorBase::emplace<FactorGnssFix3D>(ftr, ftr, sensor_gnss_, shared_from_this(), false, FAC_ACTIVE);
}

bool ProcessorGnssFix::rejectOutlier(FactorBasePtr fac)
{
    // Cast feature
    auto gnss_ftr = std::static_pointer_cast<FeatureGnssFix>(fac->getFeature());
    // copy states
    Eigen::VectorXs x(gnss_ftr->getCapture()->getFrame()->getP()->getState());
    Eigen::VectorXs o(gnss_ftr->getCapture()->getFrame()->getP()->getState());
    Eigen::VectorXs x_antena(sensor_gnss_->getStateBlock(0)->getState());
    Eigen::VectorXs t_ENU_map(sensor_gnss_->getEnuMapTranslation()->getState());
    Eigen::VectorXs roll_ENU_map(sensor_gnss_->getEnuMapRoll()->getState());
    Eigen::VectorXs pitch_ENU_map(sensor_gnss_->getEnuMapPitch()->getState());
    Eigen::VectorXs yaw_ENU_map(sensor_gnss_->getEnuMapYaw()->getState());
    // create Scalar* array of a copy of the state
    Scalar* parameters[7] = {x.data(), o.data(), x_antena.data(), t_ENU_map.data(), roll_ENU_map.data(),
                             pitch_ENU_map.data(), yaw_ENU_map.data()};
    // create residuals pointer
    Eigen::VectorXs residuals(3);
    // evaluate the factor in this state
    fac->evaluate(parameters, residuals.data(), nullptr);
    // discard if residual too high evaluated at the current estimation
    if (residuals.norm() > 1e3)
    {
        WOLF_WARN("Discarding GNSS FIX Factor, considered OUTLIER");
        WOLF_TRACE("Feature: ", fac->getMeasurement().transpose(),"\nError: ",(fac->getMeasurementSquareRootInformationUpper().inverse()*residuals).transpose());
        fac->remove();
        return true;
    }
    return false;
}

bool ProcessorGnssFix::voteForKeyFrame()
{
    // Depending on time since the last KF with gnssfix capture
    if (!last_KF_ || (incoming_capture_->getTimeStamp() - last_KF_->getTimeStamp()) > params_gnss_->time_th)
        return true;

    // TODO: more alternatives?

    // otherwise
    return false;
}

void ProcessorGnssFix::configure(SensorBasePtr _sensor)
{
    sensor_gnss_ = std::static_pointer_cast<SensorGnss>(_sensor);
};

ProcessorBasePtr ProcessorGnssFix::create(const std::string& _unique_name, const ProcessorParamsBasePtr _params, const SensorBasePtr sensor)
{
    auto prc = std::make_shared<ProcessorGnssFix>(std::static_pointer_cast<ProcessorParamsGnssFix>(_params),
                                                      std::static_pointer_cast<SensorGnss>(sensor));
    prc->setName(_unique_name);
    return prc;
}


} // namespace wolf


// Register in the SensorFactory
#include "core/processor/processor_factory.h"
namespace wolf {
WOLF_REGISTER_PROCESSOR("GNSS FIX",ProcessorGnssFix)
} // namespace wolf
