#include "gnss/sensor/sensor_gnss.h"
#include "core/state_block/state_block.h"
#include "core/state_block/state_quaternion.h"

namespace wolf {

// Geodetic system parameters
static Scalar kSemimajorAxis = 6378137;
static Scalar kSemiminorAxis = 6356752.3142;
static Scalar kFirstEccentricitySquared = 6.69437999014 * 0.001;
static Scalar kSecondEccentricitySquared = 6.73949674228 * 0.001;
//static Scalar kFlattening = 1 / 298.257223563;

SensorGnss::SensorGnss(StateBlockPtr _p_ptr,             //GNSS antena 3D position with respect to the car frame (base frame)
                       StateBlockPtr _bias_ptr,          //GNSS sensor time bias
                       IntrinsicsGnssPtr params)         //sensor params
        :
        SensorBase("GNSS", _p_ptr, nullptr, _bias_ptr, 0), // antena orientation has no sense in GNSS
        params_(params),
        ENU_defined_(false),
        ENU_MAP_initialized_(false)
{
    assert(_p_ptr->getSize() == 3 && "Bad extrinsics size");
    assert((_bias_ptr == nullptr || _bias_ptr->getSize() == 1 )&& "Bad bias size");

    getStateBlockVec().resize(7);
    setStateBlockPtrStatic(3, std::make_shared<StateBlock>(3, params_->translation_fixed_));
    setStateBlockPtrStatic(4, std::make_shared<StateBlock>(1, params_->roll_fixed_));
    setStateBlockPtrStatic(5, std::make_shared<StateBlock>(1, params_->pitch_fixed_));
    setStateBlockPtrStatic(6, std::make_shared<StateBlock>(1, params_->yaw_fixed_));
}

SensorGnss::SensorGnss(StateBlockPtr _p_ptr,             //GNSS sensor position with respect to the car frame (base frame)
                       StateBlockPtr _bias_ptr,          //GNSS sensor time bias
                       IntrinsicsGnssPtr params,         //sensor params
                       StateBlockPtr _t_ENU_MAP_ptr,     //ENU_MAP translation
                       StateBlockPtr _roll_ENU_MAP_ptr,  //ENU_MAP Roll
                       StateBlockPtr _pitch_ENU_MAP_ptr, //ENU_MAP Pitch
                       StateBlockPtr _yaw_ENU_MAP_ptr)   //ENU_MAP Yaw
        :
        SensorBase("GNSS", _p_ptr, nullptr, _bias_ptr, 0),// antena orientation has no sense in GNSS
        params_(params),
        ENU_defined_(false),
        ENU_MAP_initialized_(false)
{
    assert(_p_ptr->getSize() == 3 && "Bad extrinsics size");
    assert((_bias_ptr == nullptr || _bias_ptr->getSize() == 1 )&& "Bad bias size");
    assert(_roll_ENU_MAP_ptr->getSize() == 1 && "Bad ENU_MAP roll size");
    assert(_pitch_ENU_MAP_ptr->getSize() == 1 && "Bad ENU_MAP pitch size");
    assert(_yaw_ENU_MAP_ptr->getSize() == 1 && "Bad ENU_MAP yaw size");

    getStateBlockVec().resize(7);
    setStateBlockPtrStatic(3, _t_ENU_MAP_ptr);
    setStateBlockPtrStatic(4, _roll_ENU_MAP_ptr);
    setStateBlockPtrStatic(5, _pitch_ENU_MAP_ptr);
    setStateBlockPtrStatic(6, _yaw_ENU_MAP_ptr);
}

SensorGnss::~SensorGnss()
{
    //
}

void SensorGnss::computeEnuEcef(const Eigen::Vector3s& _t_ECEF_ENU, Eigen::Matrix3s& R_ENU_ECEF, Eigen::Vector3s& t_ENU_ECEF) const
{
    // Convert ECEF coordinates to geodetic coordinates.
    // J. Zhu, "Conversion of Earth-centered Earth-fixed coordinates
    // to geodetic coordinates," IEEE Transactions on Aerospace and
    // Electronic Systems, vol. 30, pp. 957-961, 1994.

    Scalar r = std::sqrt(_t_ECEF_ENU(0) * _t_ECEF_ENU(0) + _t_ECEF_ENU(1) * _t_ECEF_ENU(1));
    Scalar Esq = kSemimajorAxis * kSemimajorAxis - kSemiminorAxis * kSemiminorAxis;
    Scalar F = 54 * kSemiminorAxis * kSemiminorAxis * _t_ECEF_ENU(2) * _t_ECEF_ENU(2);
    Scalar G = r * r + (1 - kFirstEccentricitySquared) * _t_ECEF_ENU(2) * _t_ECEF_ENU(2) - kFirstEccentricitySquared * Esq;
    Scalar C = (kFirstEccentricitySquared * kFirstEccentricitySquared * F * r * r) / pow(G, 3);
    Scalar S = cbrt(1 + C + sqrt(C * C + 2 * C));
    Scalar P = F / (3 * pow((S + 1 / S + 1), 2) * G * G);
    Scalar Q = sqrt(1 + 2 * kFirstEccentricitySquared * kFirstEccentricitySquared * P);
    Scalar r_0 = -(P * kFirstEccentricitySquared * r) / (1 + Q)
                 + sqrt(0.5 * kSemimajorAxis * kSemimajorAxis * (1 + 1.0 / Q)
                 - P * (1 - kFirstEccentricitySquared) * _t_ECEF_ENU(2) * _t_ECEF_ENU(2) / (Q * (1 + Q)) - 0.5 * P * r * r);
    Scalar V = sqrt(pow((r - kFirstEccentricitySquared * r_0), 2) + (1 - kFirstEccentricitySquared) * _t_ECEF_ENU(2) * _t_ECEF_ENU(2));
    Scalar Z_0 = kSemiminorAxis * kSemiminorAxis * _t_ECEF_ENU(2) / (kSemimajorAxis * V);

    Scalar latitude = atan((_t_ECEF_ENU(2) + kSecondEccentricitySquared * Z_0) / r);
    Scalar longitude = atan2(_t_ECEF_ENU(1), _t_ECEF_ENU(0));

    Scalar sLat = sin(latitude);
    Scalar cLat = cos(latitude);
    Scalar sLon = sin(longitude);
    Scalar cLon = cos(longitude);

    R_ENU_ECEF(0,0) = -sLon;
    R_ENU_ECEF(0,1) =  cLon;
    R_ENU_ECEF(0,2) =  0.0;

    R_ENU_ECEF(1,0) = -sLat*cLon;
    R_ENU_ECEF(1,1) = -sLat * sLon;
    R_ENU_ECEF(1,2) =  cLat;

    R_ENU_ECEF(2,0) =  cLat * cLon;
    R_ENU_ECEF(2,1) =  cLat * sLon;
    R_ENU_ECEF(2,2) =  sLat;

    t_ENU_ECEF = -R_ENU_ECEF*_t_ECEF_ENU;
}

void SensorGnss::setEcefEnu(const Eigen::Vector3s& _t_ECEF_ENU)
{
    computeEnuEcef(_t_ECEF_ENU, R_ENU_ECEF_, t_ENU_ECEF_);
    ENU_defined_ = true;

    WOLF_INFO("SensorGnss: ECEF-ENU initialized.")
}

void SensorGnss::setEnuEcef(const Eigen::Matrix3s& _R_ENU_ECEF, const Eigen::Vector3s& _t_ENU_ECEF)
{
    R_ENU_ECEF_ = _R_ENU_ECEF;
    t_ENU_ECEF_ = _t_ENU_ECEF;
    ENU_defined_ = true;
}

void SensorGnss::initializeEnuMap(const Eigen::VectorXs& _pose_MAP_frameENU, const Eigen::Vector3s& _t_ECEF_antenaENU,
                                  const Eigen::VectorXs& _pose_MAP_frame2, const Eigen::Vector3s& _t_ECEF_antena2)
{
    Eigen::Matrix3s R_ENU_ECEF;//, R_ENU2_ECEF;
    Eigen::Vector3s t_ENU_ECEF;//, t_ENU2_ECEF;
    computeEnuEcef(_t_ECEF_antenaENU, R_ENU_ECEF, t_ENU_ECEF);
    //computeENU_ECEF(_t_ECEF_antena2, R_ENU2_ECEF, t_ENU2_ECEF);

    // compute fix vector (from 1 to 2) in ENU coordinates
    Eigen::Vector3s v_ENU = R_ENU_ECEF * (_t_ECEF_antena2 - _t_ECEF_antenaENU);

    // 2D
    if (getProblem()->getDim() == 2)
    {
        // compute antena vector (from 1 to 2) in MAP
        Eigen::Vector2s t_MAP_antenaENU = _pose_MAP_frameENU.head<2>() + Eigen::Rotation2D<Scalar>(_pose_MAP_frameENU(2)) * getP()->getState().head<2>();
        Eigen::Vector2s t_MAP_antena2 = _pose_MAP_frame2.head<2>() + Eigen::Rotation2D<Scalar>(_pose_MAP_frame2(2)) * getP()->getState().head<2>();
        Eigen::Vector2s v_MAP =  t_MAP_antena2 - t_MAP_antenaENU;

        // initialize yaw
        Scalar theta_ENU = atan2(v_ENU(1),v_ENU(0));
        Scalar theta_MAP = atan2(v_MAP(1),v_MAP(0));
        Scalar yaw_ENU_MAP = theta_ENU-theta_MAP;
        setEnuMapYawState(yaw_ENU_MAP);

        // initialize translation
        Eigen::Vector3s t_ENU_MAP(Eigen::Vector3s::Zero());
        Eigen::Matrix2s R_ENU_MAP = computeREnuMap(getEnuMapRoll()->getState()(0),
                                                   getEnuMapPitch()->getState()(0),
                                                   getEnuMapYaw()->getState()(0)).topLeftCorner<2,2>();
        t_ENU_MAP.head<2>() = -R_ENU_MAP * t_MAP_antenaENU;

        // set translation state
        setEnuMapTranslationState(t_ENU_MAP);

        //std::cout << "-----------------------------------------\n";
        //std::cout << "t_ECEF_antenaENU: " << _t_ECEF_antenaENU.transpose() << std::endl;
        //std::cout << "t_ECEF_antena2: " << _t_ECEF_antena2.transpose() << std::endl;
        //std::cout << "v_ENU: " << v_ENU.transpose() << std::endl;
        //std::cout << "theta_ENU: " << theta_ENU << std::endl;
        //std::cout << "t_MAP_antena1: " << t_MAP_antenaENU.transpose() << std::endl;
        //std::cout << "t_MAP_antena2: " << t_MAP_antena2.transpose() << std::endl;
        //std::cout << "v_MAP: " << v_MAP.transpose() << std::endl;
        //std::cout << "theta_MAP: " << theta_MAP << std::endl;
        //std::cout << "yaw set: " << yaw << std::endl;
        //std::cout << "t_ENU1_origin: " << t_ENU_MAP.transpose() << std::endl;
        //std::cout << "-----checks\n";
        //std::cout << "R(-yaw) * v_ENU: " << (Eigen::Rotation2D<Scalar>(-yaw) * v_ENU.head<2>()).transpose() << std::endl;
        //std::cout << "should be: " << v_MAP.transpose() << std::endl;
        //std::cout << "atan2(R(-yaw) * v_ENU): " << atan2((Eigen::Rotation2D<Scalar>(-yaw) * v_ENU.head<2>())(1), (Eigen::Rotation2D<Scalar>(-yaw) * v_ENU.head<2>())(0)) << std::endl;
        //std::cout << "should be: " << atan2(v_MAP(1),v_MAP(0)) << std::endl;
        //std::cout << "R(yaw) * v_MAP: " << (Eigen::Rotation2D<Scalar>(yaw) * v_MAP).transpose() << std::endl;
        //std::cout << "should be: " << v_ENU.head<2>().transpose() << std::endl;
        //std::cout << "atan2(R(yaw) * v_MAP): " << atan2((Eigen::Rotation2D<Scalar>(yaw) * v_MAP)(1),(Eigen::Rotation2D<Scalar>(yaw) * v_MAP)(0)) << std::endl;
        //std::cout << "should be: " << atan2(v_ENU(1),v_ENU(0)) << std::endl;
        //std::cout << "v_ENU.norm(): " << v_ENU.norm() << std::endl;
        //std::cout << "v_MAP.norm(): " << v_MAP.norm() << std::endl;

        WOLF_INFO("SensorGnss: ENU-MAP initialized.")
    }
    // 3D
    else
    {
        //TODO
        std::runtime_error("not implemented yet");
    }
}

void SensorGnss::initializeEnuMapYaw(const Eigen::VectorXs& _pose_MAP_frame1, const Eigen::VectorXs& _pose_MAP_frame2,
                                     const Eigen::Vector3s& _v_ECEF_antena1_antena2)
{
    assert(ENU_defined_ && "initializing ENU-MAP yaw without ENU defined");

    Eigen::Vector2s t_MAP_antena1 = _pose_MAP_frame1.head<2>() + Eigen::Rotation2D<Scalar>(_pose_MAP_frame1(2)) * getP()->getState().head<2>();
    Eigen::Vector2s t_MAP_antena2 = _pose_MAP_frame2.head<2>() + Eigen::Rotation2D<Scalar>(_pose_MAP_frame2(2)) * getP()->getState().head<2>();
    Eigen::Vector2s v_MAP_antena1_antena2 = t_MAP_antena2 - t_MAP_antena1;
    Eigen::Vector3s v_ENU_antena1_antena2 = R_ENU_ECEF_ * _v_ECEF_antena1_antena2;

    Scalar theta_ENU = atan2(v_ENU_antena1_antena2(1),v_ENU_antena1_antena2(0));
    Scalar theta_MAP = atan2(v_MAP_antena1_antena2(1),v_MAP_antena1_antena2(0));
    Scalar yaw_ENU_MAP = theta_ENU-theta_MAP;

    setEnuMapYawState(yaw_ENU_MAP);
}

void SensorGnss::setEnuMapTranslationState(const Eigen::Vector3s& t_ENU_MAP)
{
    getEnuMapTranslation()->setState(t_ENU_MAP);

    ENU_MAP_initialized_ = true;
}

void SensorGnss::setEnuMapRollState(const Scalar& roll_ENU_MAP)
{
    getEnuMapRoll()->setState(Eigen::Vector1s(roll_ENU_MAP));

    ENU_MAP_initialized_ = true;
}

void SensorGnss::setEnuMapPitchState(const Scalar& pitch_ENU_MAP)
{
    getEnuMapPitch()->setState(Eigen::Vector1s(pitch_ENU_MAP));

    ENU_MAP_initialized_ = true;
}

void SensorGnss::setEnuMapYawState(const Scalar& yaw_ENU_MAP)
{
    getEnuMapYaw()->setState(Eigen::Vector1s(yaw_ENU_MAP));

    ENU_MAP_initialized_ = true;
}

Eigen::Matrix3s SensorGnss::getREnuMap() const
{
    return Eigen::Matrix3s(Eigen::AngleAxis<Scalar>(getEnuMapRoll() ->getState()(0), Eigen::Vector3s::UnitX()) *
                           Eigen::AngleAxis<Scalar>(getEnuMapPitch()->getState()(0), Eigen::Vector3s::UnitY()) *
                           Eigen::AngleAxis<Scalar>(getEnuMapYaw()  ->getState()(0), Eigen::Vector3s::UnitZ()));
}

// Define the factory method
SensorBasePtr SensorGnss::create(const std::string& _unique_name, const Eigen::VectorXs& _extrinsics,
                              const IntrinsicsBasePtr _intrinsics)
{
    assert((_extrinsics.size() == 3 || _extrinsics.size() == 9) && "Bad extrinsics vector length. Should be 3 (antena position) or 9 (antena position and MAP-ENU initialization: position XYZ and orientation RPY)");

    SensorBasePtr sen = nullptr;
    IntrinsicsGnssPtr intrinsics_gnss_ptr = std::static_pointer_cast<IntrinsicsGnss>(_intrinsics);
    StateBlockPtr p_ptr = std::make_shared<StateBlock>(_extrinsics.head(3), intrinsics_gnss_ptr->extrinsics_fixed_);

    if (_extrinsics.size() == 3)
        sen = std::make_shared<SensorGnss>(p_ptr, nullptr, intrinsics_gnss_ptr);
    else
    {
        StateBlockPtr p_ptr             = std::make_shared<StateBlock>(_extrinsics.head<3>(),     intrinsics_gnss_ptr->extrinsics_fixed_);
        StateBlockPtr translation_ptr   = std::make_shared<StateBlock>(_extrinsics.segment<3>(3), intrinsics_gnss_ptr->translation_fixed_);
        StateBlockPtr roll_ptr          = std::make_shared<StateBlock>(_extrinsics.segment<1>(6), intrinsics_gnss_ptr->roll_fixed_);
        StateBlockPtr pitch_ptr         = std::make_shared<StateBlock>(_extrinsics.segment<1>(7), intrinsics_gnss_ptr->pitch_fixed_);
        StateBlockPtr yaw_ptr           = std::make_shared<StateBlock>(_extrinsics.segment<1>(8), intrinsics_gnss_ptr->yaw_fixed_);

        sen = std::make_shared<SensorGnss>(p_ptr, nullptr, intrinsics_gnss_ptr, translation_ptr, roll_ptr, pitch_ptr, yaw_ptr);
    }

    sen->setName(_unique_name);
    return sen;
}

} // namespace wolf


// Register in the SensorFactory
#include "core/sensor/sensor_factory.h"
namespace wolf {
WOLF_REGISTER_SENSOR("GNSS",SensorGnss)
} // namespace wolf
